/********************************************************************\
 * io.c - Functions that control the input and output of the graph. *
 *                                                                  *
 *   Input file format:                                             *
 *                                                                  *
 *   c                         Comentary line                       *
 *   s seed                    Seed used to generate the graph      *
 *   p num_vertex num_edges    Number of vertices and edges         *
 *   a i j                     Edge <i,j> (1 <= i,j <= n)           *
 *                                                                  *
 *   Output file format:                                            *
 *                                                                  *
 *   c                         Comentary line                       *
 *   m i j                     Matched edged <i,j>                  *
 *                                                                  *
 *  Obs: internaly, nodes are between 1 and n too                   *
\********************************************************************/


#include "graph.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <assert.h>


/* input_graph - read the graph from the file and build the internal
		 representation
   input       - the name of the file containing the graph information */

graph *input_graph (void) {

  FILE *input_ptr;

  int type;               /* identify the type of the line
			     in the file */
  word num_vert = 0,

       num_edges = 0,     /* Number of vertices and edges */

       edge_counter = 0;  /* verification of the number of edges */

  graph *graph_ptr;       /* pointer to the graph structure */


  if ((input_ptr = fopen (input_file_name, "r")) == NULL) {

    fprintf (stderr, "Couldn't open file (%s)\n", input_file_name);
    exit (ERROR);

  }


  graph_ptr = ALLOC (1, graph);


  /* read which type is the line */

  while ((type = fgetc (input_ptr)) != EOF) {

    switch (type) {

      case ('c'): {  /* Commentary line */

	char ch;


	do ch = fgetc (input_ptr); while (ch != '\n'); /* skip commentary */

	break;

      }

      case ('s'): {  /* seed used to generate the graph */

	word seed;


	fscanf (input_ptr, "%d\n", &seed);

	break;

      }

      case ('p'): {  /* Number of vertices and edges of the graph */

	word count;


	fscanf (input_ptr, "%d%d\n", &num_vert, &num_edges);
	if (num_vert == 0 || num_edges == 0) {

	  fprintf (stderr, "Wrong number of vertices %d or edges %d \n",
		   num_vert, num_edges);
	  exit (ERROR);

	}

	graph_ptr->num_vertices = num_vert;

	graph_ptr->num_edges = num_edges;

	/* array of vertices, vertices [1..n] / n = |V| */


	graph_ptr->vertex_ptr = ALLOC (num_vert + 1, vertex);

	/* array of mates, mate [1..n] / n = |V| */

	graph_ptr->edge_ptr = ALLOC (num_edges + 1, edge);

	add_all_vertices (graph_ptr);


	break;

      }

      case ('a'): { /* Identify an edge */

	word i, j;

#ifdef SPECIFIC_INPUT

	word match;

#endif

	if (num_vert == 0) {

	  fprintf (stderr, "Don't know how many vertices\n");
	  exit (ERROR);

	}

#ifdef SPECIFIC_INPUT

	fscanf (input_ptr, "%d%d%d\n", &i, &j, &match);

#else

	fscanf (input_ptr, "%d%d\n", &i, &j);

#endif

	if (i == 0 || j == 0) {

	  fprintf (stderr, "Edge with problem <%d, %d>\n", i, j);
	  exit (ERROR);

	}

	if (i < 1 || i > num_vert) {

	  fprintf (stderr, "Vertex %d out of range", i);
	  exit (ERROR);

	}

	if (j < 1 || j > num_vert) {

	  fprintf (stderr, "Vertex %d out of range", j);
	  exit (ERROR);

	}

	add_edge (graph_ptr, i, j);


	edge_counter++; /* verification of the number of edges */

	break;

      }

      default: {

	fprintf (stderr, "unknown type of line\n");
	exit (ERROR);

      } /* end default */

    } /* end switch */

  } /* end while not eof */

  fclose (input_ptr);

  if (edge_counter != num_edges || edge_counter == 0) {

    fprintf (stderr, "Wrong number of edges\n");
    exit (ERROR);

  }

  return graph_ptr;

}



void output_graph (graph *graph_ptr) {

  FILE *output;

  int i;


  assert ( (output = fopen (output_file_name, "w")) != NULL);

  fprintf (output, "p %d %d\n", graph_ptr->num_vertices, graph_ptr->num_edges);

  for (i = 1; i <= graph_ptr->num_edges; i++) {

    fprintf (output, "a %d %d\n",
       graph_ptr->vertex_ptr [graph_ptr->edge_ptr [i].vertex_1].new_number,
       graph_ptr->vertex_ptr [graph_ptr->edge_ptr [i].vertex_2].new_number);


  }

  fclose (output);

}
