/* io.c - Functions that control the input and output of the graph.
  
    Input file format:
 
    c                         Comentary line
    s seed                    Seed used to generate the graph
    p num_vertex num_edges    Number of vertices and edges
    a i j                     Edge <i,j> (1 <= i,j <= n)

    Output file format:

    c                         Comentary line
    m i j                     Matched edged <i,j>

   Obs: internaly, nodes are between 0 and n-1. */



#include "graph.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>


extern char input_file_name [];

extern char output_file_name [];


/* input_graph - read the graph from the file and build the internal
		 representation
   input       - the name of the file containing the graph information */

graph *input_graph (void) {

  FILE *input_ptr;

  int type;             /* identify the type of the line in the file */

  word num_vert = 0, num_edges = 0,    /* Number of vertices and edges */

       edge_counter = 0; /* verification of the number of edges */

  graph *graph_ptr;


  if ((input_ptr = fopen (input_file_name, "r")) == NULL) {

    printf ("Couldn't open file (%s)\n", input_file_name); exit (ERROR);

  }

  graph_ptr = ALLOC (1, graph);

  while ((type = fgetc (input_ptr)) != EOF) {

    switch (type) {

      case ('c'): {  /* Commentary line */

	char ch;


	do ch = fgetc (input_ptr); while (ch != '\n'); /* skip commentary */

	break;
      }

      case ('s'): {  /* seed used to generate the graph */

	int seed;


	fscanf (input_ptr, "%d\n", &seed);

	break;
      }

      case ('p'): {  /* Number of vertices and edges of the graph */

	fscanf (input_ptr, "%d%d\n", &num_vert, &num_edges);

	if (num_vert == 0 || num_edges == 0) {

	  printf ("Wrong number of vertices %d or edges %d \n",
		  num_vert, num_edges); exit (ERROR);

	}

	graph_ptr->num_vertices = num_vert;


	/* array of vertices */

	graph_ptr->vert_array = ALLOC (num_vert, vertex);


	/* array of task pointers, size |V| */

	graph_ptr->task_ptr_array = ALLOC (num_vert, task *);

	break;
      }

      case ('a'): { /* Identify an edge */

	word i, j;

#ifdef SPECIFIC_INPUT

	word match;

#endif

	if (num_vert == 0) {

	  printf ("Don't know how many vertices\n"); exit (ERROR);

	}

#ifdef SPECIFIC_INPUT

	fscanf (input_ptr, "%d%d%d\n", &i, &j, &match);

#else

	fscanf (input_ptr, "%d%d\n", &i, &j);

#endif

	if (i == 0 || j == 0) {

	  printf ("Edge with problem <%d, %d>\n", i, j); exit (ERROR);

	}

	if (i < 1 || i > num_vert) {

	  printf ("Vertex %d out of range", i); exit (ERROR);

	}

	if (j < 1 || j > num_vert) {

	  printf ("Vertex %d out of range", j); exit (ERROR);

	}

	add_edge (graph_ptr, i, j);

#ifdef SPECIFIC_INPUT

	if (match)

	  invert_edge (graph_ptr, i - 1, j - 1);

#endif

	edge_counter++; /* verification of the number of edges */

	break;
      }

      default: {

      }

    }

  }

  fclose (input_ptr);

  if (edge_counter != num_edges || edge_counter == 0) {

    printf ("Wrong number of edges\n"); exit (ERROR);

  }

  return graph_ptr;

}



/* outpur_graph - print the information of the matched edges
   input - pointer to the graph and the name of the output file
   OBS - this function alters the graph. It modifies the adjacency list. */

void output_graph (graph *graph_ptr) {

  FILE *output;

  word vert;


  if (!strcmp (output_file_name, ""))

    output = stdout;

  else

    if ((output = fopen (output_file_name, "w")) == NULL) {

      printf ("Couldn't create %s\n", output_file_name); exit (ERROR);

    }

  fprintf (output, "c matched edges\n");


  /* scan all vertices */

  for (vert = 0; vert < graph_ptr->num_vertices; vert++) {

    edge *edge_ptr = graph_ptr->vert_array [vert].edge_list;


    /* print all the edges of the vertex */

    while (edge_ptr != NULL) {

      if (edge_ptr->matched)

        if (edge_ptr->vertex != UINT_MAX) {

          edge_ptr->mate->vertex = UINT_MAX;

	  fprintf (output, "a %d %d\n", vert + 1, edge_ptr->vertex + 1);

        }

      edge_ptr = edge_ptr->next;

    }

  }

  if (strcmp (output_file_name, ""))

    fclose (output);

}


