/* library.c */


#include "graph.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <limits.h>


#ifdef UNIX

long start, end;

#else

#include <time.h>

static clock_t start, end;

#endif

char input_file_name [256];

char output_file_name [256];

static unsigned long int space_used = 0;


/* parse_command_line - interpret the command line
   input              - the command line */

void parse_command_line (int argc, char *argv []) {

  if (argc == 1 || argc > 3) {

    printf ("USAGE: match input [output]\n"); exit (NORMAL);

  }

  strcpy (input_file_name, argv [1]);

  if (argc == 3)

    strcpy (output_file_name, argv [2]);

  else

    strcpy (output_file_name, "");

}


/* alloc_mem - perform the alocation of memory, only called by
	       the define ALLOC
   output    - a pointer to a cleaned allocated memory */

void *alloc_mem (word itens, word size) {

   void *pointer;

#ifdef UNIX

   if ((pointer = calloc (itens, size)) == NULL) {

     printf ("Out of space. Space used %ld\n", space_used); exit (ERROR);

   }

#else

   if ((pointer = farcalloc (itens, size)) == NULL) {

     printf ("Out of space. Space used %ld\n", space_used); exit (ERROR);

   }

#endif

   space_used += itens * size;

   return pointer;

}


#ifdef UNIX

/* start_time - start counting the time */

void start_time (void) {

  start = clock ();

}


/* end_time - stop counting the time */

void end_time (void) {

  end = clock ();

}

#else

/* start_time - start counting the time */

void start_time (void) {

  start = clock ();

}


/* end_time - stop counting the time */

void end_time (void) {

  end = clock ();

}

#endif


/* maximal_matching_1 - produce an maximal matching in the graph
   input - a pointer to the graph structure */

void maximal_matching_1 (graph *graph_ptr) {

  word vert;


  for (vert = 0; vert < graph_ptr->num_vertices; vert++)

    if (exposed (graph_ptr, vert)) {

      edge *edge_ptr = graph_ptr->vert_array [vert].edge_list;


      while (edge_ptr != NULL) {

        if (exposed (graph_ptr, edge_ptr->vertex)) {

          edge_ptr->matched = edge_ptr->mate->matched = TRUE;

          break;

        }

        edge_ptr = edge_ptr->next;

      }

    }

}


/* cardinality - inform the number of edges incident into the vertex
   input - a pointer to the graph and the vertex */

word cardinality (graph *graph_ptr, word vert) {

  edge *edge_ptr = graph_ptr->vert_array [vert].edge_list;

  word cont = 0;


  while (edge_ptr != NULL) {

    ++cont;

    edge_ptr = edge_ptr->next;

  }

  return (cont);

}


/* compare_card - function to compere elements into the list of cardinality
   input - two pointers to elements of the list
   output - (-1,0,1) depending on the values of the elements */

int compare_card (const void *elem_a, const void *elem_b) {

  list_vert *list_vert_ptr_a = (list_vert *) elem_a,
            *list_vert_ptr_b = (list_vert *) elem_b;

  if (list_vert_ptr_a->card < list_vert_ptr_b->card)

    return (-1);

  if (list_vert_ptr_a->card > list_vert_ptr_b->card)

    return (1);

  return (0);

}


/* maximal_matching_2 - produce an maximal matching in the graph
   input - a pointer to the graph structure */

void maximal_matching_2 (graph *graph_ptr) {

  word vert, num_vert = graph_ptr->num_vertices;

  list_vert *list_vert_ptr = ALLOC (num_vert, list_vert);


  for (vert = 0; vert < num_vert; vert++) {

    list_vert_ptr [vert].card = cardinality (graph_ptr, vert);

    list_vert_ptr [vert].order = vert;

  }

  qsort ((void *) list_vert_ptr, num_vert, sizeof (list_vert), compare_card);

  for (vert = 0; vert < num_vert; vert++)

    if (exposed (graph_ptr, list_vert_ptr [vert].order)) {

      edge *edge_ptr = graph_ptr->vert_array [list_vert_ptr [vert].order].edge_list;

      while (edge_ptr != NULL) {

        if (exposed (graph_ptr, edge_ptr->vertex)) {

          edge_ptr->matched = edge_ptr->mate->matched = TRUE;

          break;

        }

        edge_ptr = edge_ptr->next;


      }

    }

#ifdef UNIX

  free (list_vert_ptr);

#else

  farfree (list_vert_ptr);

#endif

}



/* add_edge - insert an edge into the graph structure
   input    - graph, vertices i and j, (1 <= i,j <= n) */

void add_edge (graph *graph_ptr, word i, word j) {

  edge *edge_i_to_j_ptr,

       *edge_j_to_i_ptr;      /* pointers to the two new directed edges */

  vertex *vertex_ptr_i,

	 *vertex_ptr_j;


  /* create the edges */

  edge_i_to_j_ptr = ALLOC (1, edge);

  edge_j_to_i_ptr = ALLOC (1, edge);


  /* insert and link the edge <i, j> */

  vertex_ptr_i = &graph_ptr->vert_array [i - 1];

  edge_i_to_j_ptr->next = vertex_ptr_i->edge_list;

  vertex_ptr_i->edge_list = edge_i_to_j_ptr;

  edge_i_to_j_ptr->mate = edge_j_to_i_ptr;

  edge_i_to_j_ptr->vertex = j - 1;


  /* insert the edge <j, i> */

  vertex_ptr_j = &graph_ptr->vert_array [j - 1];

  edge_j_to_i_ptr->next = vertex_ptr_j->edge_list;

  vertex_ptr_j->edge_list = edge_j_to_i_ptr;

  edge_j_to_i_ptr->mate = edge_i_to_j_ptr;

  edge_j_to_i_ptr->vertex = i - 1;

}


/* add_predecessor - insert the edge into the list of the predecessors of
		     the given vertex
   input           - pointers to the vertex and to the edge */

void add_predecessor (vertex *vertex, edge *pred) {

  pred->pred_next = vertex->pred_list;

  vertex->pred_list = pred;

}


/* add_successor - insert the edge into the list of the successors of
		   the given vertex
   input         - pointers to the vertex and to the edge */

void add_successor (vertex *vertex, edge *succ) {

  succ->succ_next = vertex->succ_list;

  vertex->succ_list = succ;

}


/* add_anomaly - insert the edge into the list of the anomalies of
		 the given vertex
   input       - pointers to the vertex and to the edge */

void add_anomaly (vertex *vertex, edge *anom) {

  anom->anom_next = vertex->anom_list;

  vertex->anom_list = anom;

}


/* set_predecessor - make the list of the predecessors have only one
		     element, the given edge
   input           - pointers to the vertex and to the edge */

void set_predecessor (vertex *vertex_ptr, edge *pred_edge_ptr) {

  edge *edge_ptr = vertex_ptr->pred_list;


  /* destroy the predeccessors list of the vertex */

  while (edge_ptr != NULL) {

    edge *edge_ptr_aux = edge_ptr;


    edge_ptr = edge_ptr->pred_next;

    edge_ptr_aux->pred_next = NULL;

  }


  /* insert only the edge given */

  vertex_ptr->pred_list = pred_edge_ptr;

  pred_edge_ptr->pred_next = NULL;

}


/* set_successor - make the list of the successors have only one
		   element, the given edge
   input         - pointers to the vertex and to the edge */

void set_successor (vertex *vertex_ptr, edge *succ_edge_ptr) {

  edge *edge_ptr = vertex_ptr->succ_list;


  /* destroy the successors list of the vertex */

  while (edge_ptr != NULL) {

    edge *edge_ptr_aux = edge_ptr;


    edge_ptr = edge_ptr->succ_next;

    edge_ptr_aux->succ_next = NULL;

  }


  /* insert only the edge given */

  vertex_ptr->succ_list = succ_edge_ptr;

  succ_edge_ptr->succ_next = NULL;

}


/* set_anomaly - make the list of the anomalies have only one
		 element, the given edge
   input       - pointers to the vertex and to the edge */

void set_anomaly (vertex *vertex_ptr, edge *anom_edge_ptr) {

  edge *edge_ptr = vertex_ptr->anom_list;


  /* destroy the anomalies list of the vertex */

  while (edge_ptr != NULL) {

    edge *edge_ptr_aux = edge_ptr;


    edge_ptr = edge_ptr->anom_next;

    edge_ptr_aux->anom_next = NULL;

  }


  /* insert only the edge given */

  vertex_ptr->anom_list = anom_edge_ptr;

  anom_edge_ptr->anom_next = NULL;
}


/* base  - return the base of nested blossoms  base*(B)
   input - ponter to the graph, the initial vertex and the type of the
	   parent chain */

word base (graph *graph_ptr, word vert, DFS_Tree mark) {

 vertex *vert_ptr = &graph_ptr->vert_array [vert];


 while (vert_ptr->bloom_number != 0) {

   word vert_aux = vert_ptr->base;

   vertex *vert_aux_ptr = &graph_ptr->vert_array [vert_aux];


   /* mark the path through blossoms */

   if (mark == LEFT)

       vert_aux_ptr->left_parent = vert;

   else

       vert_aux_ptr->right_parent = vert;

   vert = vert_aux;
   vert_ptr = &graph_ptr->vert_array [vert];

 }

 return (vert);

}


/* add_bloom - insert a boom into the list of bloom in the main structure
   input     - pointer to the main structure, bloom's identification and the
	       peaks */

void add_bloom (graph *graph_ptr,
		word bloom_number,
		word vertex_left_peak, word vertex_right_peak) {

  /* alloc bloom */

  bloom *bloom_ptr = ALLOC (1, bloom);


  /* insert into the list */

  bloom_ptr->next = graph_ptr->bloom_list;

  graph_ptr->bloom_list = bloom_ptr;


  /* peaks and the identification */

  bloom_ptr->vertex_left_peak = vertex_left_peak;

  bloom_ptr->vertex_right_peak = vertex_right_peak;

  bloom_ptr->bloom_number = bloom_number;

}


/* get_bloom_peaks - retrieve the peaks from the bloom list
   input           - the pointer to the main structure, bloom's
		     identification
   output          - the peaks of the bloom */

void get_bloom_peaks (graph *graph_ptr,
		      word bloom_number,
		      word *vertex_left_peak, word *vertex_right_peak) {

  /* search the bloom */

  bloom *bloom_ptr = graph_ptr->bloom_list;

  while (bloom_ptr->bloom_number != bloom_number)

    bloom_ptr = bloom_ptr->next;


  /* retrieve the peaks */

  *vertex_left_peak = bloom_ptr->vertex_left_peak;

  *vertex_right_peak = bloom_ptr->vertex_right_peak;

}


/* add_vertex_to_bloom - insert the vertex into a list of possibly element
			 of a bloom
   input               - the list and the vertex
   output              - the new list */

bloom_creation *add_vertex_to_bloom (bloom_creation *list_ptr,
				     word vertex_num) {

 bloom_creation *node = ALLOC (1, bloom_creation);


 node->vertex_num = vertex_num;

 node->next = list_ptr;


 /* return the new first element of the list */

 return node;

}


/* create_bloom - create a new bloom that comprises all vertices marked
		  left or right during the current call of
		  blossom_augmentation
   input        - a pointer to the graph, the list of vertices used,
		  the number of the bloom and it's base. */

void create_bloom (graph *graph_ptr, bloom_creation *list_ptr,
		   word bloom_num, word base) {

  bloom_creation *bloom_creation_ptr = list_ptr;


  /* all vertices marked LEFT and RIGHT */

  while (bloom_creation_ptr != NULL) {

    vertex *vert_ptr =
	   &graph_ptr->vert_array [bloom_creation_ptr->vertex_num];


    /* skip the base of the bloom */

    if (vert_ptr->mark != NONE) {

      vert_ptr->bloom_number = bloom_num;

      vert_ptr->base = base;

      set_other_level (graph_ptr, bloom_creation_ptr->vertex_num);

    }

    bloom_creation_ptr = bloom_creation_ptr->next;

  }

}


/* free_creation_bloom_list - release the memory used by the list
   input                    - the list */

void free_creation_bloom_list (bloom_creation *list_ptr) {

  bloom_creation *bloom_creation_ptr = list_ptr;


  while (bloom_creation_ptr != NULL) {

    bloom_creation *bloom_creation_aux = bloom_creation_ptr->next;

#ifdef UNIX

    free (bloom_creation_ptr);

#else

    farfree (bloom_creation_ptr);

#endif

    bloom_creation_ptr = bloom_creation_aux;

  }

}


/* add_task - insert a task into the task list of the required phase
   input    - pointer to the graph structure, phase and a pointer to
	      the task */

void add_task (graph *graph_ptr, word phase, task *task_ptr) {

  task_ptr->next = graph_ptr->task_ptr_array [phase];

  graph_ptr->task_ptr_array [phase] = task_ptr;

}


/* verify_match - verify if the result is a valid match
   input - a pointer to the graph structure */

void verify_match (graph *graph_ptr) {

  word vert;


  /* scan all vertices */

  for (vert = 0; vert < graph_ptr->num_vertices; vert++) {

    edge *edge_ptr = graph_ptr->vert_array [vert].edge_list;

    word num_match = 0;


    /* print all the edges of the vertex */

    while (edge_ptr != NULL) {

      if (edge_ptr->matched)

	num_match++;

      edge_ptr = edge_ptr->next;

    }

    if (num_match != 0 && num_match != 1) {

      printf ("The match is incorrect\n"); exit (ERROR);

    }

  }

}


/* clear_visited_marks - clear the visited mark of the vertices and edges
   input - a pointer to the graph structure */

void clear_visited_marks (graph *graph_ptr) {

  word vert;


  /* clear all the visited mark of the the vertices */

  for (vert = 0; vert < graph_ptr->num_vertices; vert++) {

    edge *edge_ptr = graph_ptr->vert_array [vert].edge_list;


    graph_ptr->vert_array [vert].visited = FALSE;

    while (edge_ptr != NULL) {

      edge_ptr->visited = FALSE;

      edge_ptr = edge_ptr->next;

    }

  }

}


/******************** temporary */

long alive; /* used to see if the recursion of search is alive */


/* search - looks for an augmenting path starting at a ginven vertex
   input - a pointer to the graph, the vertex to start the search and
           the level of the search
   output - TRUE if an augmenting path was found */

boolean search (graph *graph_ptr, word vert, word level) {

  vertex *vert_ptr = &graph_ptr->vert_array [vert];

  edge *edge_ptr = vert_ptr->edge_list;


  if (++alive % 10000 == 0) printf ("%ld ", alive); /************ temporary */

  vert_ptr->visited = TRUE;

  if (level % 2 && level != 0) /* odd level */

    if (exposed (graph_ptr, vert)) {

      printf ("(%d)", vert + 1);

      return (TRUE);

  }

  if (!(level % 2))  /* even level */

    while (edge_ptr != NULL) {

      if (!graph_ptr->vert_array [edge_ptr->vertex].visited)

	if (search (graph_ptr, edge_ptr->vertex, level + 1)) {

          printf ("(%d)", vert + 1);

          return (TRUE);

        }

      edge_ptr = edge_ptr->next;

    }

  else  /* odd level */

    while (edge_ptr != NULL) {

      if (edge_ptr->matched)

        if (!graph_ptr->vert_array [edge_ptr->vertex].visited)

          if (search (graph_ptr, edge_ptr->vertex, level + 1)) {

            printf ("(%d)", vert + 1);

            return (TRUE);

          }

      edge_ptr = edge_ptr->next;

    }

  vert_ptr->visited = FALSE;

  return (FALSE);

}


/* verify_maximum_match - partially verify the maximum match
   input - a pointer to the graph structure */

void verify_maximum_match (graph *graph_ptr) {

  word vert;


  alive = 0; /* ******************* temporary */

  clear_visited_marks (graph_ptr);

  for (vert = 0; vert < graph_ptr->num_vertices; vert++)

    if (exposed (graph_ptr, vert)) {

      printf ("\nSearch (%d) ", vert + 1); /************** temporary */

      if (search (graph_ptr, vert, 0)) {

	printf ("There is an alternating path, match not maximum\n");

        exit (ERROR);

      }

    }

}


/* print_graph - output all the information stored into the graph structure
   input       - pointer to the graph structure
   OBS - this function alters the adjacency list of the graph */

void print_graph (graph *graph_ptr) {

  word vert;

  word matched_edges = 0;


  printf ("Number of vertices %d\n", graph_ptr->num_vertices);

  clear_visited_marks (graph_ptr);


  /* print all the edges */

  for (vert = 0; vert < graph_ptr->num_vertices; vert++) {

    edge *edge_ptr = graph_ptr->vert_array [vert].edge_list;


    /* print all the edges of the vertex */

    while (edge_ptr != NULL) {

      if (!edge_ptr->visited) {

        edge_ptr->mate->visited = TRUE;

        printf ("<%d, %d> ", vert + 1, edge_ptr->vertex + 1);

        if (edge_ptr->matched) {

	  printf (" matched \n");

	  matched_edges++;

        }

        else

	  printf (" free \n");

      }

      edge_ptr = edge_ptr->next;

    }

  }

  printf ("Matched edges %d\n", matched_edges);

  clear_visited_marks (graph_ptr);

}


/* statistc_graph - output all the information stored into the graph
		    structure
   input          - pointer to the graph structure */

void statistic_graph (graph *graph_ptr) {

  word vert;

  word matched_edges = 0;

  word free_vert = 0;

  word edges = 0;

  boolean vert_free;


  printf ("Vert %d ", graph_ptr->num_vertices);


  /* print all the vertex informations, including edges. */

  for (vert = 0; vert < graph_ptr->num_vertices; vert++) {

    edge *edge_ptr = graph_ptr->vert_array [vert].edge_list;

    vert_free = TRUE;


    /* print all the edges of the vertex */

    while (edge_ptr != NULL) {

      edges++;

      if (edge_ptr->matched) {

        vert_free = FALSE;

	matched_edges++;

      }

      edge_ptr = edge_ptr->next;

    }

    if (vert_free)

      free_vert++;

  }

#ifdef UNIX

   printf ("M %d T %f Mem %lu\n",
	  matched_edges / 2, (float) end / 1000000.0, space_used); 
/* printf ("Edges %d Max match %d Free %d Time %f, Mem req %lu\n",
	  edges / 2, matched_edges / 2, free_vert,
	  (float) end / 1000000.0, space_used); 
*/
#else

/*  printf ("Edges %d Max match %d Free %d\nTime %f Mem left %lu, Mem req %lu\n",
	  edges / 2, matched_edges / 2, free_vert,
	  (end - start) / CLK_TCK,
	   farcoreleft (), space_used);
*/

printf ("M %d\n", matched_edges /2);

#endif

  return;

}


/* free_graph - release all the allocated memory used by the graph structure
   input      - pointer to the graph structure */

void free_graph (graph *graph_ptr) {

  bloom *bloom_ptr;

  word counter = 0;


  /* free all the edges in the graph */

  for (counter = 0; counter < graph_ptr->num_vertices; counter++) {

    edge *edge_ptr = graph_ptr->vert_array [counter].edge_list;


    while (edge_ptr != NULL) {

      edge *edge_aux = edge_ptr;


      edge_ptr = edge_ptr->next;

#ifdef UNIX

      free (edge_aux);

#else

      farfree (edge_aux);

#endif

    }

  }

#ifdef UNIX

  free (graph_ptr->vert_array);

#else

  farfree (graph_ptr->vert_array);

#endif

  bloom_ptr = graph_ptr->bloom_list;

  while (bloom_ptr != NULL) {

    bloom *bloom_aux = bloom_ptr;


    bloom_ptr = bloom_ptr->next;

#ifdef UNIX

    free (bloom_aux);

#else

    farfree (bloom_aux);

#endif

  }


  /* free all tasks in all phases */

  for (counter = 0; counter < graph_ptr->num_vertices; counter++) {

    task *task_ptr = graph_ptr->task_ptr_array [counter];


    while (task_ptr != NULL) {

      task *task_aux = task_ptr;


      task_ptr = task_ptr->next;

#ifdef UNIX

      free (task_aux);

#else

      farfree (task_aux);

#endif

    }

  }

#ifdef UNIX

  free (graph_ptr->task_ptr_array);

  free (graph_ptr);

#else

  farfree (graph_ptr->task_ptr_array);

  farfree (graph_ptr);

#endif

}


