/* matching.c - solve the matching problem on generic graphs */


#include "graph.h"

#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <limits.h>


boolean augmentation [2]; /* true if an augmentation occurr
			     augmentation [0] (phase - 1)
			     augmentation [1] (actual phase) */
word phase;               /* number or the actual phase */

word bloom_number;




/* start_aug_info - initialize the information of the augmentation occurred
		    at the current and the anterior */

void start_aug_info (void) {

  augmentation [0] = augmentation [1] = FALSE;

}


/* next_phase_aug - set the information of the augmentation to start
		    the next phase */

void next_phase_aug (void) {

  augmentation [0] = augmentation [1];

  augmentation [1] = FALSE;

}


/* actual_phase_aub - inform if an augmentation occurred at the current
		      phase.
   output - TRUE if an augmentation occurred, FALSE otherwise */

boolean actual_phase_aug (void) {

  return augmentation [1];

}


/* anterior_phase_aub - inform if an augmentation occurred at the anterior
			phase.
   output - TRUE if an augmentation occurred, FALSE otherwise */

boolean anterior_phase_aug (void) {

  return augmentation [0];

}


/* set_atual_phase_aug - inform that an augmentation occurred at the current
			 phase */

void set_actual_phase_aug (void) {

  augmentation [1] = TRUE;

}


/* initiatialization - initialize all fields in the graph to start the
		       next phase
   input - pointer to the main structure */

void initialization (graph *graph_ptr) {

  word vert;

  word phase;


  for (vert = 0; vert < graph_ptr->num_vertices; vert++) {

    edge *edge_ptr;

    vertex *vert_ptr = &graph_ptr->vert_array [vert];


    /* levels with infinity */

    vert_ptr->even_level = vert_ptr->odd_level = UINT_MAX;


    /* bloom to which vertex belongs, 0 => bloom undefined */

    vert_ptr->bloom_number = 0;

    vert_ptr->base = 0;

    vert_ptr->count = 0;

    vert_ptr->left_parent = vert_ptr->right_parent = vert;

    vert_ptr->erased = FALSE;

    vert_ptr->visited = FALSE;

    vert_ptr->mark = NONE;

    vert_ptr->pred_list = vert_ptr->succ_list = vert_ptr->anom_list = NULL;


    /* initialize the edges */

    edge_ptr = graph_ptr->vert_array [vert].edge_list;

    while (edge_ptr != NULL) {

      edge_ptr->used = edge_ptr->visited = FALSE;

      edge_ptr->pred_next = edge_ptr->succ_next = edge_ptr->anom_next = NULL;

      edge_ptr = edge_ptr->next;

    }

  }


  /* initialize the task */

  for (phase = 0; phase < graph_ptr->num_vertices; phase++) {

    task *task_ptr = graph_ptr->task_ptr_array [phase];


    while (task_ptr != NULL) {

      task *task_aux_ptr = task_ptr;


      task_ptr = task_ptr->next;

#ifdef UNIX

      free (task_aux_ptr);

#else

      farfree (task_aux_ptr);

#endif

    }

    graph_ptr->task_ptr_array [phase] = NULL;

  }

}


/* level - inform the level of a vertex, min {odd_level, even_level}
   input - pointer to the graph and the vertex index
   output - level of the vertex */

word level (graph *graph_ptr, word vertex) {

  word odd, even;

  return ((odd = graph_ptr->vert_array [vertex].odd_level) <
	  (even = graph_ptr->vert_array [vertex].even_level) ? odd : even);

}


/* exposed - inform if the vertex has no incident matched edge.
   input - graph pointer and the vertex index
   output - TRUE if the vertex is exposed, FALSE otherwise */

boolean exposed (graph *graph_ptr, word vertex) {

  edge *edge_ptr;


  edge_ptr = graph_ptr->vert_array [vertex].edge_list;

  while (edge_ptr != NULL) {

    if (edge_ptr->matched)

      return (FALSE);

    edge_ptr = edge_ptr->next;

  }

  return (TRUE);

}


/* initialize_tasks - scan all the exposed vertex and create the initial
		      tasks (breadth-first search tasks)
   input - pointer to the graph structure */

boolean initialize_tasks (graph *graph_ptr) {

  word vert;

  boolean there_is_exposed = FALSE;


  for (vert = 0; vert < graph_ptr->num_vertices; vert++) {

    if (exposed (graph_ptr, vert)) {

      task *task_ptr = ALLOC (1, task);

      there_is_exposed = TRUE;

      graph_ptr->vert_array [vert].even_level = 0;
      task_ptr->type = BFS;
      task_ptr->vert_1 = vert;

      add_task (graph_ptr, phase, task_ptr);

    }

  }

  return (there_is_exposed);

}


/* there_is_tasks - inform if there is tasks to be done at the current phase
   input - graph pointer
   output - TRUE if exist tasks, FALSE otherwise */

boolean there_is_tasks (graph *graph_ptr) {

  return (graph_ptr->task_ptr_array [phase] != NULL ? TRUE : FALSE);

}


/* get_task - get a task from the task list of the current phase
   input - pointer to the graph structure
   output - pointer to the task in memory, NULL if there is no talk */

task *get_task (graph *graph_ptr) {

  task *task_ptr;


  if ((task_ptr = graph_ptr->task_ptr_array [phase]) != NULL) {

    graph_ptr->task_ptr_array [phase] =
    graph_ptr->task_ptr_array [phase]->next;

    task_ptr->next = NULL;

  }

  return task_ptr;

}


/* even_BFS - execute a step of the breadth-first search from the specified
	      vertex creating tasks to the current phase if a bridge is
	      discovered and to the next phase if the search must continue
   input - graph pointer and the vertex index */

void even_BFS (graph *graph_ptr, word vertex_v) {

  vertex *vertex_ptr_v = &graph_ptr->vert_array [vertex_v];

  edge *edge_ptr = graph_ptr->vert_array [vertex_v].edge_list;


  while (edge_ptr != NULL) { /* for each unerased neighbor of vertex
				such that the edge is free */
    if (!edge_ptr->matched) { /* edge is free */

      word vertex_u = edge_ptr->vertex;

      vertex *vertex_ptr_u = &graph_ptr->vert_array [vertex_u];


      if (!vertex_ptr_u->erased)


	/* unerased */

	if (vertex_ptr_u->even_level < UINT_MAX) { /* less than infinity */


	  /* bridge */

	  word j = (vertex_ptr_v->even_level + vertex_ptr_u->even_level) / 2;

	  task *task_ptr = ALLOC (1, task);


	  task_ptr->type = BRIDGE;

	  task_ptr->vert_1 = vertex_v;
	  task_ptr->vert_2 = vertex_u;

	  add_task (graph_ptr, j, task_ptr);

	} /* endif bridge found */

	else {

	  if (vertex_ptr_u->odd_level == UINT_MAX) /* odd_level = infinity */

	    vertex_ptr_u->odd_level = phase + 1;


	  if (vertex_ptr_u->odd_level == phase + 1) {

	    task *task_ptr = ALLOC (1, task);

	    vertex_ptr_u->count++;


	    /* insert v into predecessors (u) */

	    add_predecessor (vertex_ptr_u, edge_ptr->mate);


	    /* insert u into predecessors (v) */

	    add_successor (vertex_ptr_v, edge_ptr);

	    task_ptr->type = BFS;

	    task_ptr->vert_1 = vertex_u;

	    add_task (graph_ptr, phase + 1, task_ptr);

	  }

	  if (vertex_ptr_u->odd_level < phase)

	    add_anomaly (vertex_ptr_u, edge_ptr->mate);

	} /* endif expand bread-first search */

    } /* endif umatched edge */

    edge_ptr = edge_ptr->next;

  } /* whileend all neighbors */

}



/* odd_BFS - execute a step of the breadth-first search from the specified
	     vertex creating tasks to the current phase if a bridge is
	     discovered and to the next phase if the search must continue
   input - graph pointer and the vertex index */

void odd_BFS (graph *graph_ptr, word vertex_v) {

  vertex *vertex_ptr_v = &graph_ptr->vert_array [vertex_v];

  edge *edge_ptr = graph_ptr->vert_array [vertex_v].edge_list;

  word vertex_u;

  vertex *vertex_ptr_u;


  if (vertex_ptr_v->bloom_number != 0) /* vertex_v belong to a bloom */

    return;


  /* let vertex_u be the mate of vertex_v */

  while (!edge_ptr->matched)

    edge_ptr = edge_ptr->next;

  vertex_u = edge_ptr->vertex;

  vertex_ptr_u = &graph_ptr->vert_array [vertex_u];

  if (vertex_ptr_u->odd_level < UINT_MAX) { /* new bridge */

    word j = (vertex_ptr_u->odd_level + vertex_ptr_v->odd_level) / 2;

    task *task_ptr = ALLOC (1, task);


    task_ptr->type = BRIDGE;

    task_ptr->vert_1 = vertex_v;
    task_ptr->vert_2 = vertex_u;

    add_task (graph_ptr, j, task_ptr);

    return;

  }

  if (vertex_ptr_u->even_level == UINT_MAX) { /* infinity */

    task *task_ptr = ALLOC (1, task);


    /* predecessors (u) := {v} */

    set_predecessor (vertex_ptr_u, edge_ptr->mate);


    /* successors (v) := {u} */

    set_successor (vertex_ptr_v, edge_ptr);


    vertex_ptr_u->count = 1;

    vertex_ptr_u->even_level = phase + 1;

    task_ptr->type = BFS;

    task_ptr->vert_1 = vertex_u;

    add_task (graph_ptr, phase + 1, task_ptr);

  }

}


/* left_DFS - execute one step of the double depth-first search to find
	       the exposed vertices.
   input - pointer to the graph structure, the peak s, the current
	   vertex of the left and right search and the indicative if
	   the right searche fails.
   output - TRUE if a bloom is discovered, the DCV and the list of all
	    vertices visited by the two searches, right and left */

boolean left_DFS (graph *graph_ptr,
		  word vertex_s,
		  word *vertex_left, word *DCV, word *vertex_right,
		  bloom_creation **list_ptr,
		  boolean *right_fail) {

  edge *edge_ptr = graph_ptr->vert_array [*vertex_left].pred_list;


  while (edge_ptr != NULL) {

    word vertex_u = edge_ptr->vertex;

    vertex *vertex_ptr_u = &graph_ptr->vert_array [vertex_u];


    /* choose an "unused" predecessor edge (vertex_left, u) such that
       u is "unerased */

    if (!edge_ptr->used && !vertex_ptr_u->erased) {

      edge_ptr->used = edge_ptr->mate->used = TRUE;

      vertex_ptr_u->left_parent = *vertex_left;


      /* skip blooms */

      if (vertex_ptr_u->bloom_number != 0) {

	vertex_u = base (graph_ptr, vertex_u, LEFT);
	vertex_ptr_u = &graph_ptr->vert_array [vertex_u];

      }


      /* mark vertex if possible */

      if (vertex_ptr_u->mark == NONE) {

	vertex_ptr_u->mark = LEFT;

	*list_ptr = add_vertex_to_bloom (*list_ptr, vertex_u);

	*vertex_left = vertex_u;

	if (level (graph_ptr, *vertex_left) <=
	    level (graph_ptr, *vertex_right))

	  *right_fail = FALSE;

	return (FALSE);

      }

    } /* endif unerased predecessor */

    edge_ptr=edge_ptr->pred_next;

  }

  if (*vertex_left == vertex_s)

    if (*right_fail)

      return (TRUE);

    else { /* make the left search act as the right search */

      *DCV = *vertex_left = *vertex_right;

      graph_ptr->vert_array [*vertex_left].mark = LEFT;

      *vertex_right = graph_ptr->vert_array [*vertex_right].right_parent;

    }

  else

    *vertex_left = graph_ptr->vert_array [*vertex_left].left_parent;

  return (FALSE);

}


/* right_DFS - execute one step of the double depth-first search to find
	       the exposed vertices.
   input - pointer to the graph structure, the center of activities of the
	   right and left search, the DCV, barrier
   output - the DCV, barrier, the indicative of failure of right search
	    and the list of vertices visited by the search */

void right_DFS (graph *graph_ptr,
		word *vertex_left, word *vertex_right,
		word *DCV, word *barrier, bloom_creation **list_ptr,
		boolean *right_fail) {

  edge *edge_ptr = graph_ptr->vert_array [*vertex_right].pred_list;


  while (edge_ptr != NULL) {

    word vertex_u = edge_ptr->vertex;

    vertex *vertex_ptr_u = &graph_ptr->vert_array [vertex_u];


    /* choose an "unused" predecessor edge (vertex_left, u) such that
       u is "unerased */

    if (!edge_ptr->used && !vertex_ptr_u->erased) {

      edge_ptr->used = edge_ptr->mate->used = TRUE;

      vertex_ptr_u->right_parent = *vertex_right;


      /* skip blooms */

      if (vertex_ptr_u->bloom_number != 0) {

	vertex_u = base (graph_ptr, vertex_u, RIGHT);
	vertex_ptr_u = &graph_ptr->vert_array [vertex_u];

      }


      /* mark vertex if possible */

      if (vertex_ptr_u->mark == NONE) {

	vertex_ptr_u->mark = RIGHT;

	*list_ptr = add_vertex_to_bloom (*list_ptr, vertex_u);

	*vertex_right = vertex_u;

	return;

      }

      else

	if (vertex_u == *vertex_left)

	  *DCV = vertex_u;

    }

    edge_ptr = edge_ptr->pred_next;

  }

  if (*vertex_right == *barrier) {

    *right_fail = TRUE;

    *vertex_right = *DCV;

    *barrier = *DCV;

    graph_ptr->vert_array [*vertex_right].mark = RIGHT;

    *vertex_left = graph_ptr->vert_array [*vertex_left].left_parent;

  }

  else

    *vertex_right = graph_ptr->vert_array [*vertex_right].right_parent;

}


/* reverse_path - invert the direction of a given path 
   input - the begining and the end of the path
   ouput - the inverted path */

void reverse_path (find **find_begin, find **find_end) {

  find *find_aux = *find_begin;


  do {

    find *find_tmp = find_aux->next;


    find_aux->next = find_aux->ant;
    find_aux->ant = find_tmp;
    find_aux = find_aux->ant; /* next is now ant */

  } while (find_aux != NULL);

  find_aux = *find_begin;
  *find_begin = *find_end;
  *find_end = find_aux;

}


/* open_path - find path through blooms
   input - pointer to the graph structure and the vertex that belongs
	   to a bloom 
   output - the pointers of the begining and the end of the path */

void open_path (graph *graph_ptr,
		word vertex_x,
		find **find_begin, find **find_end) {

  word bloom_number = graph_ptr->vert_array [vertex_x].bloom_number;

  word base = graph_ptr->vert_array [vertex_x].base;


  if (level (graph_ptr, vertex_x) % 2) { /* inner */

    find *path_1_begin, *path_1_end,
	 *path_2_begin, *path_2_end;

    word left_peak, right_peak;


    get_bloom_peaks (graph_ptr, bloom_number, &left_peak, &right_peak);

    if (graph_ptr->vert_array [vertex_x].mark == LEFT) {

      find_path (graph_ptr,
		 left_peak, vertex_x,
		 bloom_number, LEFT,
		 &path_1_begin, &path_1_end);

      find_path (graph_ptr,
		 right_peak, base,
		 bloom_number, RIGHT,
		 &path_2_begin, &path_2_end);

    }

    else {

      find_path (graph_ptr,
		 right_peak, vertex_x,
		 bloom_number, RIGHT,
		 &path_1_begin, &path_1_end);

      find_path (graph_ptr,
		 left_peak, base,
		 bloom_number, LEFT,
		 &path_2_begin, &path_2_end);

    }

    reverse_path (&path_1_begin, &path_1_end);

    path_2_begin->ant = path_1_end;
    path_1_end->next = path_2_begin;

    *find_begin = path_1_begin;
    *find_end = path_2_end;

  }

  else /* outer */

    /* Modification based on R.BRUCE MATTINGLY AND NATHAN P. RITCHEY */
    find_path (graph_ptr,
	       vertex_x, base,
	       bloom_number, NONE,
	       find_begin, find_end);

}


/* find_path - find the path from high to low
   input - graph pointer, the vertices high and low a bloom identification
   output - the pointers of the begining and the end of the path */

void find_path (graph *graph_ptr,
		word vertex_high, word vertex_low,
		word bloom_number,
		DFS_Tree mark,
		find **find_begin, find **find_end) {

  word vertex_u;
  vertex *vertex_ptr_u;

  edge *edge_ptr;

  find *find_actual;
  vertex *vert_ptr_act;

  find *find_aux;


  *find_begin = find_actual = ALLOC (1, find);

  find_actual->vert_num = vertex_high;

  if (vertex_high == vertex_low) {

    *find_end = *find_begin;

    return;

  }

  vert_ptr_act = &graph_ptr->vert_array [find_actual->vert_num];

  edge_ptr = vert_ptr_act->pred_list;

  do {

    while (TRUE) { /* while vertex_v has no unvisited predecessor edges */

      if (edge_ptr == NULL) {

	do {

	  find_aux = find_actual;

	  find_actual = find_actual->ant;

	  vert_ptr_act = &graph_ptr->vert_array [find_actual->vert_num];

#ifdef UNIX

	  free (find_aux);

#else

	  farfree (find_aux);

#endif

	} while (vert_ptr_act->bloom_number != 0 &&
		 vert_ptr_act->bloom_number != bloom_number &&
		 find_actual->vert_num != vertex_high);

	edge_ptr = vert_ptr_act->pred_list;

      }

      if (!edge_ptr->visited)

	break;

      edge_ptr = edge_ptr->pred_next;

    }


    /* am I into a bloom ? */

    if (vert_ptr_act->bloom_number == 0 ||
	vert_ptr_act->bloom_number == bloom_number) {

      edge_ptr->visited = TRUE;
      edge_ptr->mate->visited = TRUE;

      vertex_u = edge_ptr->vertex;
      vertex_ptr_u = &graph_ptr->vert_array [vertex_u];

    }

    else {

      vertex_u = vert_ptr_act->base;
      vertex_ptr_u = &graph_ptr->vert_array [vertex_u];

    }


    /* visit vertex if possible */

    if (!vertex_ptr_u->visited &&
	!vertex_ptr_u->erased  &&
	level (graph_ptr, vertex_u) > level (graph_ptr, vertex_low) &&
	((vertex_ptr_u->mark == mark) ||
	 (vertex_ptr_u->bloom_number != 0 &&
	  vertex_ptr_u->bloom_number != bloom_number) ||
	 (mark == NONE))) {

      find_aux = ALLOC (1, find);


      /* mark vertices only outside of blooms */

      if (vertex_ptr_u->bloom_number == 0 ||
	  vertex_ptr_u->bloom_number == bloom_number)
	vertex_ptr_u->visited = TRUE;

      find_aux->vert_num = vertex_u;

      find_aux->ant = find_actual;
      find_actual->next = find_aux;
      find_actual = find_aux;

      vert_ptr_act = &graph_ptr->vert_array [find_actual->vert_num];

      edge_ptr = vert_ptr_act->pred_list;

    }

    else

      edge_ptr = edge_ptr->pred_next;

  } while (vertex_u != vertex_low);


  /* the path begins with vertex_u */

  find_aux = ALLOC (1, find);

  find_aux->vert_num = vertex_u;

  find_aux->ant = find_actual;
  find_actual->next = find_aux;
  find_actual = find_aux;

  *find_end = find_actual;


  /* go through the path to open blooms */

  do {

    vertex *vert_ptr_aux;


    find_aux = find_aux->ant;

    vert_ptr_aux = &graph_ptr->vert_array [find_aux->vert_num];

    if ((vert_ptr_aux->bloom_number != 0) &&
	(vert_ptr_aux->bloom_number != bloom_number)) {

      find *find_aux_begin, *find_aux_end;

      find *find_tmp;


      open_path (graph_ptr,
		 find_aux->vert_num,
		 &find_aux_begin, &find_aux_end);


      /* insert path into the original */

      if (find_aux->ant != NULL) { /* It isn't the begining of the path */ 

	find_aux_begin->ant = find_aux->ant;

	find_aux->ant->next = find_aux_begin;

      }

      if (find_aux->next->next != NULL) { /* it's not the end of the path */

	find_aux_end->next = find_aux->next->next;

	find_aux->next->next->ant = find_aux_end;

      }

      find_tmp = find_aux;

      if (find_tmp->ant != NULL) /* continue the search */

	find_aux = find_aux_begin;

      else /* insert the new begining of the path */

	*find_begin = find_aux = find_aux_begin;

      if (find_tmp->next->next == NULL) /* insert the new end of the path */

	*find_end = find_aux_end;

#ifdef UNIX

      free (find_tmp->next);

      free (find_tmp);

#else

      farfree (find_tmp->next);

      farfree (find_tmp);

#endif

    }

  } while (find_aux->vert_num != vertex_high);

}


/* invert_edge - makes the edge between the two vertices switch between
		 matched and unmatched.
   input - graph pointer and two adjacent vertices. */

void invert_edge (graph *graph_ptr, word vert_u, word vert_v) {

  vertex *vert_u_ptr = &graph_ptr->vert_array [vert_u];

  edge *edge_ptr = vert_u_ptr->edge_list;


  while (edge_ptr->vertex != vert_v)

    edge_ptr = edge_ptr->next;

  if (edge_ptr->matched)

    edge_ptr->matched = edge_ptr->mate->matched = FALSE;

  else

    edge_ptr->matched = edge_ptr->mate->matched = TRUE;

}


/* invert_path - switch the edges of the path. The matched edges become
		 unmatched and vice versa.
   input - graph pointer and the begining of the path */

void invert_path (graph *graph_ptr, find *find_begin) {

  find *find_aux = find_begin;


  while (find_aux->next != NULL) {

    invert_edge (graph_ptr, find_aux->vert_num, find_aux->next->vert_num);

    find_aux = find_aux->next;

  }

}


/* erase_successors - scan all the successors subtracting the number of
		      unerased predeccessors. If it becomes 0, than the
		      vertex is erased too.
   input - graph pointer and a vertex */

void erase_successors (graph *graph_ptr, word vert) {

  vertex *vert_ptr = &graph_ptr->vert_array [vert];

  edge *edge_ptr = vert_ptr->succ_list;


  while (edge_ptr != NULL) {

    vertex *vertex_z_ptr = &graph_ptr->vert_array [edge_ptr->vertex];


    if (!vertex_z_ptr->erased)

      if (--vertex_z_ptr->count == 0) {

	vertex_z_ptr->erased = TRUE;

	erase_successors (graph_ptr, edge_ptr->vertex);

      }

    edge_ptr = edge_ptr->succ_next;

  }

}


/* erase_path - erase all the vertices belongin to the path between the
		vertices high and low.
   input - graph pointer and the pointers to the begining and the end of the
	   path */

void erase_path (graph *graph_ptr, find *find_begin) {

  find *find_aux = find_begin;


  while (find_aux != NULL) {

    vertex *vert_ptr_aux = &graph_ptr->vert_array [find_aux->vert_num];


    vert_ptr_aux->erased = TRUE;

    erase_successors (graph_ptr, find_aux->vert_num);

    find_aux = find_aux->next;

  }

}


/* free_path - release the memory allocated to the path 
   input - the begining of the path */

void free_path (find *find_begin) {

  find *find_ptr = find_begin;


  while (find_ptr != NULL) {

    find *find_aux = find_ptr;


    find_ptr = find_ptr->next;

#ifdef UNIX

    free (find_aux);

#else

    farfree (find_aux);

#endif

  }

}


/* set_other_level - compute the other level of the given vertex and
		     search the anomalies to newly bridges
   input - graph pointer and the vertex */

void set_other_level (graph *graph_ptr, word vertex_y) {

  vertex *vertex_ptr_y = &graph_ptr->vert_array [vertex_y];


  if (level (graph_ptr, vertex_y) % 2) { /* inner */

    edge *edge_ptr;

    task *task_ptr = ALLOC (1, task);


    vertex_ptr_y->even_level = 2 * phase + 1 - vertex_ptr_y->odd_level;

    task_ptr->type = BFS;
    task_ptr->vert_1 = vertex_y;
    add_task (graph_ptr, vertex_ptr_y->even_level, task_ptr);

    edge_ptr = vertex_ptr_y->anom_list;

    while (edge_ptr != NULL) {

      word j = (vertex_ptr_y->even_level +
		graph_ptr->vert_array [edge_ptr->vertex].even_level) / 2;


      task_ptr = ALLOC (1, task);

      task_ptr->type = BRIDGE;
      task_ptr->vert_1 = vertex_y;
      task_ptr->vert_2 = edge_ptr->vertex;
      add_task (graph_ptr, j, task_ptr);

      edge_ptr->used = TRUE;
      edge_ptr->mate->used = TRUE;

      edge_ptr = edge_ptr->anom_next;

    }

  }

  else /* outer */

    vertex_ptr_y->odd_level = 2 * phase + 1 - vertex_ptr_y->even_level;

}


/* blossom_augmentation - either discovers a bloom or augments the current
			 matching
   input - graph pointer and the peaks of a bridge s and t */

void blossom_augmentation (graph *graph_ptr, word vertex_s, word vertex_t) {

  vertex *vertex_ptr_s = &graph_ptr->vert_array [vertex_s];
  vertex *vertex_ptr_t = &graph_ptr->vert_array [vertex_t];

  word vertex_left,
       vertex_right;

  word DCV,     /* Deepest commom vertex */
       barrier;

  task *task_ptr;

  bloom_creation *bloom_creation_list = NULL;

  boolean right_fail;


  /* if s and t are both unerased the continue */

  if (vertex_ptr_s->erased || vertex_ptr_t->erased)
    return;


  /* if s and t belong to the same bloom then return */

  if (vertex_ptr_s->bloom_number == vertex_ptr_t->bloom_number &&
      vertex_ptr_s->bloom_number != 0)
    return;


  /* if s belongs to a bloom */

  if (vertex_ptr_s->bloom_number != 0)

    vertex_left = base (graph_ptr, vertex_s, LEFT);

  else

    vertex_left = vertex_s;


  /* if v belong to a bloom */

  if (vertex_ptr_t->bloom_number != 0)

    vertex_right = base (graph_ptr, vertex_t, RIGHT);

  else

    vertex_right = vertex_t;


  /* Modification based on R.BRUCE MATTINGLY AND NATHAN P. RITCHEY */

  if (vertex_right == vertex_left)

    return;

  graph_ptr->vert_array [vertex_left].mark = LEFT;

  bloom_creation_list = add_vertex_to_bloom (bloom_creation_list, vertex_left);

  graph_ptr->vert_array [vertex_right].mark = RIGHT;

  bloom_creation_list = add_vertex_to_bloom (bloom_creation_list, vertex_right);

  DCV = UINT_MAX; /* DCV = undefined */

  barrier = vertex_t;

  right_fail = FALSE;


  /* while both ar not exposed */

  while (!exposed (graph_ptr, vertex_left) ||
	 !exposed (graph_ptr, vertex_right)) {

    boolean bloom_discovered = FALSE;


    if (level (graph_ptr, vertex_left) >=
	level (graph_ptr, vertex_right))

      bloom_discovered = left_DFS (graph_ptr, vertex_s, &vertex_left, &DCV,
				   &vertex_right, &bloom_creation_list,
				   &right_fail);

    else

      right_DFS (graph_ptr, &vertex_left, &vertex_right,
		 &DCV, &barrier, &bloom_creation_list, &right_fail);

    if (bloom_discovered) {

      graph_ptr->vert_array [DCV].mark = NONE;

      add_bloom (graph_ptr, ++bloom_number, vertex_s, vertex_t);

      create_bloom (graph_ptr, bloom_creation_list, bloom_number, DCV);

      free_creation_bloom_list (bloom_creation_list);

      return;

    }

  }

  task_ptr = ALLOC (1, task);

  task_ptr->type = FINDPATH;

  task_ptr->vert_1 = vertex_s;
  task_ptr->vert_2 = vertex_left;
  task_ptr->path_1 = TO_BE_DONE;

  task_ptr->vert_3 = vertex_t;
  task_ptr->vert_4 = vertex_right;
  task_ptr->path_2 = TO_BE_DONE;

  add_task (graph_ptr, phase, task_ptr);

  free_creation_bloom_list (bloom_creation_list);

  set_actual_phase_aug ();

}


/* task_manager - control the distribution and execution of tasks
   input - graph pointer */

void task_manager (graph *graph_ptr) {

  task *task_ptr;

  find *path_1_begin, *path_1_end,
       *path_2_begin, *path_2_end;

  phase = 0;


/*  if (! */
   initialize_tasks (graph_ptr);
/*) {

    phase++;

    next_phase_aug ();

    printf ("no more exposed vertices\n");

    return;

  }
*/
  while (there_is_tasks (graph_ptr) && !anterior_phase_aug ()) {

    while ((task_ptr = get_task (graph_ptr)) != NULL) {
							 
      switch (task_ptr->type) {

	case BFS: {

	  if (phase % 2) /* odd level */

	    odd_BFS (graph_ptr, task_ptr->vert_1);

	  else           /* even level */

	    even_BFS (graph_ptr, task_ptr->vert_1);

#ifdef UNIX

	  free (task_ptr);

#else

	  farfree (task_ptr);

#endif

	  break;

	}

	case BRIDGE: {

	  blossom_augmentation (graph_ptr, task_ptr->vert_1, task_ptr->vert_2);

#ifdef UNIX

	  free (task_ptr);

#else

	  farfree (task_ptr);

#endif

	  break;

	}


	case FINDPATH: {

	  if (task_ptr->path_1 == TO_BE_DONE) {

	    task_ptr->path_1 = BEING_DONE;

	    add_task (graph_ptr, phase, task_ptr);

	    find_path (graph_ptr,
		       task_ptr->vert_1, task_ptr->vert_2,
		       0, LEFT,
		       &path_1_begin, &path_1_end);

	    invert_path (graph_ptr, path_1_begin);

	    task_ptr->path_1 = DONE;

	  }

	  else

	    if (task_ptr->path_2 == TO_BE_DONE) {

	      task_ptr->path_2 = BEING_DONE;

	      add_task (graph_ptr, phase, task_ptr);

	      find_path (graph_ptr,
			 task_ptr->vert_3, task_ptr->vert_4,
			 0, RIGHT,
			 &path_2_begin, &path_2_end);

	      invert_path (graph_ptr, path_2_begin);

	      task_ptr->path_2 = DONE;

	    }

	    else

	      if (task_ptr->path_1 == DONE &&
		  task_ptr->path_2 == DONE) {

		invert_edge (graph_ptr, task_ptr->vert_1, task_ptr->vert_3);

		erase_path (graph_ptr, path_1_begin);

		erase_path (graph_ptr, path_2_begin);

		free_path (path_1_begin);

		free_path (path_2_begin);

#ifdef UNIX

		free (task_ptr);

#else

		farfree (task_ptr);

#endif
	      }

	  break;

	}

      }

    }

    phase++;

    next_phase_aug ();

  }

}


/* general_matching_algorithm - main loop of the algorithm
   input - graph pointer */

void general_matching_algorithm (graph *graph_ptr) {

  do {

    start_aug_info ();

    initialization (graph_ptr);

    task_manager (graph_ptr);

  } while (anterior_phase_aug ()); /* until no augmentation occurred */

}
